/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "houdini.pmh"

#define FE_SCH_DEBUG	FALSE

namespace fe
{

SurfaceCurvesHoudini::SurfaceCurvesHoudini(void):
	m_pGdp(NULL),
	m_subIndex(-1)
{
#if FE_SCH_DEBUG
	feLog("SurfaceCurvesHoudini()\n");
#endif
}

SurfaceCurvesHoudini::~SurfaceCurvesHoudini(void)
{
#if FE_SCH_DEBUG
	feLog("~SurfaceCurvesHoudini()\n");
#endif
}

Protectable* SurfaceCurvesHoudini::clone(Protectable* pInstance)
{
#if FE_SCH_DEBUG
	feLog("SurfaceCurvesHoudini::clone\n");
#endif

	SurfaceCurvesHoudini* pSurfaceCurvesHoudini= pInstance?
			dynamic_cast<SurfaceCurvesHoudini*>(pInstance):
			new SurfaceCurvesHoudini();

	SurfaceCurves::clone(pSurfaceCurvesHoudini);

	// TODO copy attributes

	return pSurfaceCurvesHoudini;
}

void SurfaceCurvesHoudini::cache(void)
{
#if FE_SCH_DEBUG
	feLog("SurfaceCurvesHoudini::cache\n");
#endif

	//* NOTE not doing any Record serialization

	//* convert Houdini data to vertex and normal arrays

	if(!m_pGdp)
	{
		feLog("SurfaceCurvesHoudini::cache NULL gdp\n");
		return;
	}

#ifdef FE_HOUDINI_USE_GA
	const GA_PrimitiveGroup* pPrimitiveGroup=NULL;
#else
	GB_PrimitiveGroup* pPrimitiveGroup=NULL;
#endif
	if(!m_group.empty())
	{
		pPrimitiveGroup=m_pGdp->findPrimitiveGroup(m_group.c_str());
	}

#ifdef FE_HOUDINI_USE_GA
	const I32 entries=pPrimitiveGroup?
			pPrimitiveGroup->entries(): m_pGdp->getNumPrimitives();
#else
	const I32 entries=pPrimitiveGroup?
			pPrimitiveGroup->entries(): m_pGdp->primitives().entries();
#endif

	if(m_subIndex>=entries)
	{
		feLog("SurfaceCurvesHoudini::cache excessive subIndex %d/%d\n",
				m_subIndex,entries);
		return;
	}
	const U32 count=(m_subIndex>=0)? 1: entries;

#if FE_SCH_DEBUG
	feLog("SurfaceCurvesHoudini::cache count %d\n",count);
#endif

	m_vertices=0;
	U32 allocated=0;

	GEO_AttributeHandle handleNormal=
			m_pGdp->getAttribute(GEO_POINT_DICT,GEO_STD_ATTRIB_NORMAL);
	GEO_AttributeHandle handlePartition=
			m_pGdp->getPrimAttribute(m_partitionAttr.c_str());

	const GEO_Primitive* pGroupPrimitive=NULL;
	U32 total=0;
	for(U32 m=0;m<count;m++)
	{
		if(pPrimitiveGroup)
		{
			if(!m)
			{
				pGroupPrimitive=m_pGdp->primitives().head(*pPrimitiveGroup);
				if(m_subIndex>=0)
				{
					for(U32 n=0;n<m_subIndex;n++)
					{
						FEASSERT(pGroupPrimitive);
						pGroupPrimitive=m_pGdp->primitives().next(
								pGroupPrimitive,*pPrimitiveGroup);
					}
				}
			}
			else
			{
				FEASSERT(pGroupPrimitive);
				pGroupPrimitive=m_pGdp->primitives().next(
						pGroupPrimitive,*pPrimitiveGroup);
			}
		}

		const U32 primitiveIndex=(m_subIndex>=0)? m_subIndex: m;

#ifdef FE_HOUDINI_USE_GA
		const GEO_Primitive* pPrimitive=pPrimitiveGroup?
				pGroupPrimitive: m_pGdp->getGEOPrimitive(
				m_pGdp->primitiveOffset(primitiveIndex));
#else
		const GEO_Primitive* pPrimitive=pPrimitiveGroup?
				pGroupPrimitive: m_pGdp->primitives()(primitiveIndex);
#endif

		if(!pPrimitive)
		{
			break;
		}

		I32 partitionIndex= -1;
		handlePartition.setElement(pPrimitive);
		if(handlePartition.isAttributeValid())
		{
			UT_String utString;
			handlePartition.getString(utString,0);
			const String partition=utString.buffer();
			partitionIndex=lookupPartition(partition);
		}

#ifdef FE_HOUDINI_USE_GA
		if(pPrimitive->getTypeId().get()==GEO_PRIMPOLY)
#else
		if(pPrimitive->getPrimitiveId()==GEOPRIMPOLY)
#endif
		{
#if FE_SCH_DEBUG
			feLog("SurfaceCurvesHoudini::cache prim poly\n");
#endif

			const U32 subCount=pPrimitive->getVertexCount();
			if(subCount<2)
			{
				feLog("SurfaceCurvesHoudini::cache as GEO_PrimPoly"
						" primitive %d/%d subCount=%d (not triangles?)\n",
						primitiveIndex,count,subCount);
				continue;
			}

			m_vertices+=subCount;

			U32 desired=allocated;
			if(!allocated)
			{
				desired=m_vertices*count;
			}
			else if(allocated<m_vertices)
			{
				desired=m_vertices*1.2;	//* tweak
			}
			if(allocated!=desired)
			{
#if FE_SCH_DEBUG
				feLog("resize %d\n",desired);
#endif
				resizeArrays(desired);
				allocated=desired;
			}

			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				FEASSERT(total<allocated);

				//* location
				UT_Vector4 utVector4=
						pPrimitive->getVertex(subIndex).getPt()->getPos();
				m_pVertexArray[total]=SpatialVector(utVector4.data());

				//* normal
				const U32 pointIndex=
						pPrimitive->getVertex(subIndex).getBasePt()->getNum();
#ifdef FE_HOUDINI_USE_GA
				const GEO_Point* pPoint=
						m_pGdp->getGEOPoint(m_pGdp->pointOffset(pointIndex));
#else
				const GEO_Point* pPoint=m_pGdp->points()(pointIndex);
#endif
				handleNormal.setElement(pPoint);
				utVector4=handleNormal.getV4(0);
				m_pNormalArray[total]=SpatialVector(utVector4.data());
				m_pTriangleIndexArray[total]=primitiveIndex;
				m_pPrimitiveIndexArray[total]=primitiveIndex;
				m_pPartitionIndexArray[total]=partitionIndex;
				set(m_pUVArray[total],subIndex/(subCount-1.0),0.0);

#if FE_SCH_DEBUG
				feLog("poly prim %d/%d sub %d/%d part %d uv %s\n",
						primitiveIndex,count,
						subIndex,subCount,partitionIndex,
						c_print(m_pUVArray[total]));
				feLog("  vert %s norm %s\n",
						c_print(m_pVertexArray[total]),
						c_print(m_pNormalArray[total]));
#endif


				total++;
			}
		}

#ifdef FE_HOUDINI_USE_GA
		if(pPrimitive->getTypeId().get()==GEO_PRIMNURBSURF ||
				pPrimitive->getTypeId().get()==GEO_PRIMNURBCURVE)
#else
		if(pPrimitive->getPrimitiveId()==GEOPRIMNURBSURF ||
				pPrimitive->getPrimitiveId()==GEOPRIMNURBCURVE)
#endif
		{
			feX(e_unsupported,"SurfaceCurvesHoudini::cache",
					"searchable NURBS curves not yet supported");

			//* TODO
		}

#ifdef FE_HOUDINI_USE_GA
		if(pPrimitive->getTypeId().get()==GEO_PRIMBEZSURF ||
				pPrimitive->getTypeId().get()==GEO_PRIMBEZCURVE)
#else
		if(pPrimitive->getPrimitiveId()==GEOPRIMBEZSURF ||
				pPrimitive->getPrimitiveId()==GEOPRIMBEZCURVE)
#endif
		{
			feX(e_unsupported,"SurfaceCurvesHoudini::cache",
					"searchable Bezier curves not yet supported");

			//* TODO
		}
	}
	if(allocated!=m_vertices)
	{
		resizeArrays(m_vertices);
		allocated=m_vertices;
	}

	if(total!=m_vertices)
	{
		feLog("SurfaceCurvesHoudini::cache converted %d/%d vertices\n",
				total,m_vertices);
		for(U32 m=total;m<m_vertices;m++)
		{
			set(m_pVertexArray[m]);
			set(m_pNormalArray[m],0.0,1.0,0.0);
			set(m_pUVArray[m]);
			m_pTriangleIndexArray[m]= -1;
			m_pPrimitiveIndexArray[total]= -1;
			m_pPartitionIndexArray[total]= -1;
		}
	}

#if FE_SCH_DEBUG
	feLog("SurfaceCurvesHoudini::cache %d vertices\n",m_vertices);
#endif

	calcBoundingSphere();
}

} // namespace
